import pytest

from yasiu_native.decorators import flexible_decorator, flexible_decorator_2d


def testDecorOneLevel_1():

    @flexible_decorator
    def decor1(fun):
        def wrapper(*a, **kw):
            return fun(*a, **kw)
        return wrapper

    @decor1
    def funTest():
        pass
    funTest()


def testDecorOneLevel_2():

    @flexible_decorator
    def decor1(fun):
        def wrapper(*a, **kw):
            return fun(*a, **kw)
        return wrapper

    @decor1()
    def funTest():
        pass
    funTest()


def testDecorOneLevel_3():

    @flexible_decorator
    def decor1(fun, a=2):
        def wrapper(*a, **kw):
            return fun(*a, **kw)
        assert a == 5, "True"
        return wrapper

    @decor1(5)
    def funTest():
        pass
    funTest()


def testDecorOneLevel_4():

    @flexible_decorator
    def decor1(fun, a, b, g):
        def wrapper(*a, **kw):
            return fun(*a, **kw)
        return wrapper

    @decor1(2, 3, g=10)
    def funTest():
        pass
    funTest()


def testDecorOneLevel_5():
    "Test of decoration with function"

    @flexible_decorator
    def decor1(fun, callThis, a, b, g):
        def wrapper(*a, **kw):
            callThis(a, b)
            return fun(*a, **kw)
        return wrapper

    @decor1(sum, 2, 3, g=10)
    def funTest():
        pass
    funTest()


def testDecorOneLevel_6():
    "Test class decorators"

    @flexible_decorator
    def decor1(fun):
        def wrapper(*a, **kw):
            return fun(*a, **kw)
        return wrapper

    class TClass:
        @decor1
        def prt(self):
            pass

        @decor1
        @staticmethod
        def prtST(cls):
            pass

    ob = TClass()
    ob.prt()

    TClass.prtST


def testDecorOneLevel_7():
    "Test class decorators"

    @flexible_decorator
    def decor1(fun, param1=3):

        assert param1 == 5, "True"

        def wrapper(*a, **kw):
            return fun(*a, **kw)
        return wrapper

    class TClass:
        @decor1(5)
        def prt(self):
            pass

        @decor1(5)
        @staticmethod
        def prtST(cls):
            pass

    ob = TClass()
    ob.prt()

    TClass.prtST


# def testDecorOneLevel_8():
#     "Test of decoration with function"

#     @flexible_decorator
#     def decor1(fun, callThis):
#         def wrapper(*a, **kw):
#             print(f"Call this: {callThis}")
#             # callThis(1, 2)
#             return fun(*a, **kw)
#         return wrapper

#     @decor1(sum)
#     def funTest():
#         pass
#     funTest()


"""

=== Two Level Test ===

"""


def testDecorTwoLevel_1():

    @flexible_decorator_2d
    def decor2():
        def wrapper(fun):
            def inner(*a, **kw):
                return fun(*a, **kw)
            return inner
        return wrapper

    @decor2
    def testFun():
        pass

    testFun()

    @decor2
    def testFun(a=2):
        pass

    testFun()


def testDecorTwoLevel_2():

    @flexible_decorator_2d
    def decor2():
        def wrapper(fun):
            def inner(*a, **kw):
                return fun(*a, **kw)
            return inner
        return wrapper

    @decor2()
    def testFun():
        pass

    testFun()


def testDecorTwoLevel_3():

    @flexible_decorator_2d
    def decor2(param=2):

        assert param == 5, "True"

        def wrapper(fun):
            def inner(*a, **kw):
                return fun(*a, **kw)
            return inner
        return wrapper

    @decor2(param=5)
    def testFun():
        pass

    testFun()

    @decor2(5)
    def testFun():
        pass


def testDecorTwoLevel_3():

    @flexible_decorator_2d
    def decor2(param=2):

        assert param == 5, "True"

        def wrapper(fun):
            def inner(*a, **kw):
                return fun(*a, **kw)
            return inner
        return wrapper

    @decor2(param=5)
    def testFun():
        pass

    testFun()

    @decor2(5)
    def testFun():
        pass

    testFun()
    testFun()



from yasiu_native.decorators import flexible_decorator_2d as _flex_dec_2d


@_flex_dec_2d
def measure_perf_time_decorator(fmt=">4.1f"):
    """
    Example:
        @measure_perf_time_decorator()
        def func():
            ...

        @measure_perf_time_decorator(">4.1f")
        def func():
            ...

    Args:
        fmt: string format (PEP 3101)

    Returns:

    """
    from functools import wraps
    import time

    def decorator(fun):
        @wraps(wrapped=fun)
        def wrapper(*a, **kw):
            # fmt = ">4.1f"
            time0 = time.perf_counter()
            res = fun(*a, **kw)
            t_end = time.perf_counter()
            dur = t_end - time0
            if dur < 1e-3:
                timeend = f"{dur * 1000000:{fmt}} us"
            elif dur < 1:
                timeend = f"{dur * 1000:{fmt}} ms"
            elif dur > 90:
                "Minutes"
                timeend = f"{dur / 60:{fmt}} min"
            elif dur > (60*60):
                "Hours"
                timeend = f"{dur / 60 / 60:{fmt}} h"
            else:
                timeend = f"{dur:{fmt}} s"
            print(f"{fun.__name__} perf exec time: {timeend}.")

            return res

        return wrapper

    return decorator


@_flex_dec_2d
def measure_real_time_decorator(fmt=">4.1f"):
    """
    Example:
        @measure_real_time_decorator()
        def func():
            ...

        @measure_real_time_decorator(">4.1f")
        def func():
            ...

    Args:
        fmt: string format (PEP 3101)

    Returns:

    """
    from functools import wraps
    import time

    def decorator(fun):
        @wraps(wrapped=fun)
        def wrapper(*a, **kw):
            # fmt = ">4.1f"
            time0 = time.time()
            res = fun(*a, **kw)
            t_end = time.time()
            dur = t_end - time0
            if dur < 1e-3:
                timeend = f"{dur * 1000000:{fmt}} us"
            elif dur < 1:
                timeend = f"{dur * 1000:{fmt}} ms"
            elif dur > 90:
                "Minutes"
                timeend = f"{dur / 60:{fmt}} min"
            elif dur > (60*60):
                "Hours"
                timeend = f"{dur / 60 / 60:{fmt}} h"
            else:
                timeend = f"{dur:{fmt}} s"
            print(f"{fun.__name__} real exec time: {timeend}.")

            return res

        return wrapper

    return decorator


__all__ = ['measure_perf_time_decorator', 'measure_real_time_decorator']

if __name__ == "__main__":
    pass



from functools import update_wrapper as _update_wrapper


def flexible_decorator(decorator):
    """
    Decorator for decorators.
    Turns Single Level decorator into two leveled decorator,
    by passing decoration parameters alongside function signature for ease of use.
    Decorators wrapped with `Flexible` can be used both with `()` and without `()` operator.
    Child decorator receive function signature, decoration parameters and function arguments.
    Supports both positional and key arguments on any level.

    Args:
        decorator - decorator for wrapping

    Returns:
        your decorated decorator: for decorating other functions.

    Warning:
        Do not pass functions as decoration parameter as FIRST positional argument!

        @decor1(sum)
        def funTest():
            "Wrong Decoration will happen!"
            pass

    Example:

          @flexible_decorator
          def yourDecorator(decoratedFunction, *posParam, **keyParam):
            "decoratedFunction: function To be decorated with your decorator."
            "posParam: [Optional] positional argument for customizng your decorator behaviour"
            "keyParam: [Optional] keyword argument for customizng your decorator behaviour"

            def inner(*args, **kwargs):
                "args kwargs: arguments of decorated funciton"
                "Use decorative arguments to modify decorator"

                ret = decoratedFunction(*args, **kwargs)
                return ret
            return inner

    Usage:

        yourDecorator
        def someFunction(a=1, b=2):
            pass

        yourDecorator()
        def someFunction(a=1, b=2):
            pass

        yourDecorator(decorationParam)
        def someFunction(a=1, b=2):
            pass

        yourDecorator(decorationParam=5)
        def someFunction(a=1, b=2):
            pass
    """
    def wrapper(*args, **kw):
        if len(args) == 1:
            "If more arguments, then it is not function reference"
            fun1 = args[0]
        else:
            fun1 = None

        if callable(fun1):
            "Decorated without calling () in decoration"
            "No arguments were used during"
            decorWrapped = _update_wrapper(decorator, fun1)
            return decorWrapped(fun1)

        def inner(fun2, ):
            ""
            decorWrapped = _update_wrapper(decorator, fun2)
            ret = decorWrapped(fun2, *args, **kw)  # (*a2, **kw2)
            return ret

        return inner

    return wrapper


def flexible_decorator_2d(decorator):
    """
    Decorator for decorators made of 2 function levels.
    Decorators wrapped with `Flexible` can be used both with `()` and without `()` operator.
    Child decorator receive function signature, decoration parameters and function arguments.
    Supports both positional and key arguments on any level.

    Args:
        decorator - decorator for wrapping

    Returns:
        your decorated decorator: for decorating other functions.

    Warning:
        Do not pass functions as decoration parameter as FIRST positional argument!

        @decor1(sum)
        def funTest():
            "Wrong Decoration will happen!"
            pass

    Example:

          @flexible_decorator_2d
          def yourDecorator(*posParam, **keyParam):
            "posParam: [Optional] positional argument for customizng your decorator behaviour"
            "keyParam: [Optional] keyword argument for customizng your decorator behaviour"

            def wrapper(decoratedFunction):
                "decoratedFunction: function To be decorated with your decorator."

                def inner(*args, **kwargs):
                    "args kwargs: arguments of decorated funciton"
                    "Use decorative arguments to modify decorator"

                    ret = decoratedFunction(*args, **kwargs)
                    return ret

                return inner

            return wrapper

    Usage:

        yourDecorator
        def someFunction(a=1, b=2):
            pass

        yourDecorator()
        def someFunction(a=1, b=2):
            pass

        yourDecorator(decorationParam)
        def someFunction(a=1, b=2):
            pass

        yourDecorator(decorationParam=5)
        def someFunction(a=1, b=2):
            pass
    """

    def wrapper(*args, **kw):
        if len(args) == 1:
            fun1 = args[0]
        else:
            fun1 = None

        if callable(fun1):
            "Decorated without calling () in decoration"
            "No arguments were used during"
            decorWrapped = _update_wrapper(decorator, fun1)
            return decorWrapped()(fun1)

        def inner(fun2, ):
            decorWrapped = _update_wrapper(decorator, fun2)
            ret = decorWrapped(*args, **kw)(fun2)

            return ret

        return inner

    return wrapper


__all__ = ['flexible_decorator', 'flexible_decorator_2d']

if __name__ == "__main__":
    print("\n"*1)

    @flexible_decorator_2d
    def decor1(fun, param=5):
        "DEcor1 Docstring your decor"
        def wrapper(asd):
            def inner(*a, **kw):
                print(f"decor1 called, param:{asd}")
                return fun(*a, **kw)
            return inner
        return wrapper

    @decor1()
    def test1(a, b=2, c=3):
        print(f"a:{a}, b:{b}, c:{c}")

    test1(0, 1, 2)
    test1(10)
